@@ -8,7 +8,9 @@
  */
 
 /*
- * clipboard.c: Functions to handle the clipboard
+ * clipboard.c: Functions to handle the clipboard. Additionally contains the
+ *		clipboard provider code, which is separate from the main
+ *		clipboard code.
  */
 
 #include "vim.h"
@@ -29,6 +31,11 @@
 // versions of these for the 'clipboard' selection, as Visual mode has no use
 // for them.
 
+
+#ifdef FEAT_CLIPBOARD_PROVIDER
+static int clip_provider_is_available(char_u *provider);
+#endif
+
 #if defined(FEAT_CLIPBOARD)
 
 #if defined(FEAT_WAYLAND_CLIPBOARD)
@@ -100,10 +107,6 @@ typedef struct {
 
     clip_wl_selection_T regular;
     clip_wl_selection_T primary;
-
-    // Array of file descriptors of clients we are sending data to. These should
-    // be polled for POLLOUT and have the respective callback called for each.
-    garray_T write_fds;
 } clip_wl_T;
 
 // Mime types we support sending and receiving
@@ -134,12 +137,6 @@ static bool clip_wl_owner_exists(Clipboard_T *cbd);
 
 #endif // FEAT_WAYLAND_CLIPBOARD
 
-#ifdef FEAT_CLIPBOARD_PROVIDER
-static int clip_provider_is_available(Clipboard_T *cbd, char_u *provider);
-static void clip_provider_set_selection(Clipboard_T *cbd, char_u *provider);
-static void clip_provider_request_selection(Clipboard_T *cbd, char_u *provider);
-#endif
-
 /*
  * Selection stuff using Visual mode, for cutting and pasting text to other
  * windows.
@@ -178,22 +175,6 @@ skip:
     }
 }
 
-    static void
-clip_init_single(Clipboard_T *cb, int can_use)
-{
-    // No need to init again if cbd is already available
-    if (can_use && cb->available)
-	return;
-
-    cb->available  = can_use;
-    cb->owned      = FALSE;
-    cb->start.lnum = 0;
-    cb->start.col  = 0;
-    cb->end.lnum   = 0;
-    cb->end.col    = 0;
-    cb->state      = SELECT_CLEARED;
-}
-
 /*
  * Check whether the VIsual area has changed, and if so try to become the owner
  * of the selection, and free any old converted selection we may still have
@@ -237,36 +218,32 @@ clip_update_selection(Clipboard_T *clip)
 }
 
     static int
-clip_gen_own_selection(Clipboard_T *cbd UNUSED)
+clip_gen_own_selection(Clipboard_T *cbd)
 {
-    if (clipmethod == CLIPMETHOD_GUI)
-    {
-#ifdef FEAT_GUI
-	if (gui.in_use)
-	    return clip_mch_own_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_WAYLAND)
-    {
-#ifdef FEAT_WAYLAND_CLIPBOARD
-	return clip_wl_own_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_X11)
-    {
-#ifdef FEAT_XCLIPBOARD
-	return clip_xterm_own_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_OTHER)
-    {
-#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
-	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
-	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
+#if defined(FEAT_XCLIPBOARD) || defined(FEAT_WAYLAND_CLIPBOARD)
+# ifdef FEAT_GUI
+    if (gui.in_use)
 	return clip_mch_own_selection(cbd);
+    else
+# endif
+    {
+	if (clipmethod == CLIPMETHOD_WAYLAND)
+	{
+#ifdef FEAT_WAYLAND_CLIPBOARD
+	    return clip_wl_own_selection(cbd);
 #endif
+	}
+	else if (clipmethod == CLIPMETHOD_X11)
+	{
+#ifdef FEAT_XCLIPBOARD
+	    return clip_xterm_own_selection(cbd);
+#endif
+	}
     }
     return FAIL;
+#else
+    return clip_mch_own_selection(cbd);
+#endif
 }
 
     void
@@ -306,35 +283,31 @@ clip_own_selection(Clipboard_T *cbd)
 }
 
     static void
-clip_gen_lose_selection(Clipboard_T *cbd UNUSED)
+clip_gen_lose_selection(Clipboard_T *cbd)
 {
-    if (clipmethod == CLIPMETHOD_GUI)
-    {
-#ifdef FEAT_GUI
-	if (gui.in_use)
-	    clip_mch_lose_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_WAYLAND)
-    {
-#ifdef FEAT_WAYLAND_CLIPBOARD
-	clip_wl_lose_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_X11)
-    {
-#ifdef FEAT_XCLIPBOARD
-	clip_xterm_lose_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_OTHER)
-    {
-#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
-	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
-	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
+#if defined(FEAT_XCLIPBOARD) || defined(FEAT_WAYLAND_CLIPBOARD)
+# ifdef FEAT_GUI
+    if (gui.in_use)
 	clip_mch_lose_selection(cbd);
+    else
+# endif
+    {
+	if (clipmethod == CLIPMETHOD_WAYLAND)
+	{
+#ifdef FEAT_WAYLAND_CLIPBOARD
+	    clip_wl_lose_selection(cbd);
 #endif
+	}
+	else if (clipmethod == CLIPMETHOD_X11)
+	{
+#ifdef FEAT_XCLIPBOARD
+	    clip_xterm_lose_selection(cbd);
+#endif
+	}
     }
+#else
+    clip_mch_lose_selection(cbd);
+#endif
 }
 
     void
@@ -1360,77 +1333,57 @@ clip_gen_set_selection(Clipboard_T *cbd)
 	    return;
 	}
     }
-    if (clipmethod == CLIPMETHOD_GUI)
-    {
-#ifdef FEAT_GUI
-	if (gui.in_use)
-	    clip_mch_set_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_WAYLAND)
-    {
-#ifdef FEAT_WAYLAND_CLIPBOARD
-	clip_wl_set_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_X11)
-    {
-#ifdef FEAT_XCLIPBOARD
-	clip_xterm_set_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_OTHER)
-    {
-#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
-	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
-	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
+#if defined(FEAT_XCLIPBOARD) || defined(FEAT_WAYLAND_CLIPBOARD)
+# ifdef FEAT_GUI
+    if (gui.in_use)
 	clip_mch_set_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_PROVIDER)
+    else
+# endif
     {
-#ifdef FEAT_CLIPBOARD_PROVIDER
-	clip_provider_set_selection(cbd, clipprovider_name);
+	if (clipmethod == CLIPMETHOD_WAYLAND)
+	{
+#ifdef FEAT_WAYLAND_CLIPBOARD
+	    clip_wl_set_selection(cbd);
 #endif
+	}
+	else if (clipmethod == CLIPMETHOD_X11)
+	{
+#ifdef FEAT_XCLIPBOARD
+	    clip_xterm_set_selection(cbd);
+#endif
+	}
     }
+#else
+    clip_mch_set_selection(cbd);
+#endif
 }
 
     static void
-clip_gen_request_selection(Clipboard_T *cbd UNUSED)
+clip_gen_request_selection(Clipboard_T *cbd)
 {
-    if (clipmethod == CLIPMETHOD_GUI)
-    {
+#if defined(FEAT_XCLIPBOARD) || defined(FEAT_WAYLAND_CLIPBOARD)
 # ifdef FEAT_GUI
-	if (gui.in_use)
-	    clip_mch_request_selection(cbd);
-# endif
-    }
-    else if (clipmethod == CLIPMETHOD_WAYLAND)
-    {
-#ifdef FEAT_WAYLAND_CLIPBOARD
-	clip_wl_request_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_X11)
-    {
-#ifdef FEAT_XCLIPBOARD
-	clip_xterm_request_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_OTHER)
-    {
-#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
-	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
-	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
+    if (gui.in_use)
 	clip_mch_request_selection(cbd);
-#endif
-    }
-    else if (clipmethod == CLIPMETHOD_PROVIDER)
+    else
+# endif
     {
-#ifdef FEAT_CLIPBOARD_PROVIDER
-	clip_provider_request_selection(cbd, clipprovider_name);
+	if (clipmethod == CLIPMETHOD_WAYLAND)
+	{
+#ifdef FEAT_WAYLAND_CLIPBOARD
+	    clip_wl_request_selection(cbd);
 #endif
+	}
+	else if (clipmethod == CLIPMETHOD_X11)
+	{
+#ifdef FEAT_XCLIPBOARD
+	    clip_xterm_request_selection(cbd);
+#endif
+	}
     }
+#else
+    clip_mch_request_selection(cbd);
+#endif
 }
 
 #if (defined(FEAT_X11) && defined(FEAT_XCLIPBOARD) && defined(USE_SYSTEM)) \
@@ -1447,128 +1400,33 @@ clip_x11_owner_exists(Clipboard_T *cbd)
     int
 clip_gen_owner_exists(Clipboard_T *cbd UNUSED)
 {
-    if (clipmethod == CLIPMETHOD_OTHER)
-    {
+#if defined(FEAT_XCLIPBOARD) || defined(FEAT_WAYLAND_CLIPBOARD)
 # ifdef FEAT_GUI_GTK
-	if (gui.in_use)
-	    return clip_gtk_owner_exists(cbd);
-# endif
-    }
-    else if (clipmethod == CLIPMETHOD_WAYLAND)
-    {
-# ifdef FEAT_WAYLAND_CLIPBOARD
-	return clip_wl_owner_exists(cbd);
-# endif
-    }
-    else if (clipmethod == CLIPMETHOD_X11)
-    {
-# ifdef FEAT_XCLIPBOARD
-	return clip_x11_owner_exists(cbd);
-# endif
-    }
+    if (gui.in_use)
+	return clip_gtk_owner_exists(cbd);
     else
-	return FALSE;
-    return FALSE;
-}
-#endif
-
-/*
- * Extract the items in the 'clipboard' option and set global values.
- * Return an error message or NULL for success.
- */
-    char *
-did_set_clipboard(optset_T *args UNUSED)
-{
-    int		new_unnamed = 0;
-    int		new_autoselect_star = FALSE;
-    int		new_autoselect_plus = FALSE;
-    int		new_autoselectml = FALSE;
-    int		new_html = FALSE;
-    regprog_T	*new_exclude_prog = NULL;
-    char	*errmsg = NULL;
-    char_u	*p;
-
-    for (p = p_cb; *p != NUL; )
+# endif
     {
-	// Note: Keep this in sync with p_cb_values.
-	if (STRNCMP(p, "unnamed", 7) == 0 && (p[7] == ',' || p[7] == NUL))
+	if (clipmethod == CLIPMETHOD_WAYLAND)
 	{
-	    new_unnamed |= CLIP_UNNAMED;
-	    p += 7;
+#ifdef FEAT_WAYLAND_CLIPBOARD
+	    return clip_wl_owner_exists(cbd);
+#endif
 	}
-	else if (STRNCMP(p, "unnamedplus", 11) == 0
-					    && (p[11] == ',' || p[11] == NUL))
+	else if (clipmethod == CLIPMETHOD_X11)
 	{
-	    new_unnamed |= CLIP_UNNAMED_PLUS;
-	    p += 11;
-	}
-	else if (STRNCMP(p, "autoselect", 10) == 0
-					    && (p[10] == ',' || p[10] == NUL))
-	{
-	    new_autoselect_star = TRUE;
-	    p += 10;
-	}
-	else if (STRNCMP(p, "autoselectplus", 14) == 0
-					    && (p[14] == ',' || p[14] == NUL))
-	{
-	    new_autoselect_plus = TRUE;
-	    p += 14;
-	}
-	else if (STRNCMP(p, "autoselectml", 12) == 0
-					    && (p[12] == ',' || p[12] == NUL))
-	{
-	    new_autoselectml = TRUE;
-	    p += 12;
-	}
-	else if (STRNCMP(p, "html", 4) == 0 && (p[4] == ',' || p[4] == NUL))
-	{
-	    new_html = TRUE;
-	    p += 4;
-	}
-	else if (STRNCMP(p, "exclude:", 8) == 0 && new_exclude_prog == NULL)
-	{
-	    p += 8;
-	    new_exclude_prog = vim_regcomp(p, RE_MAGIC);
-	    if (new_exclude_prog == NULL)
-		errmsg = e_invalid_argument;
-	    break;
+#ifdef FEAT_XCLIPBOARD
+	    return clip_x11_owner_exists(cbd);
+#endif
 	}
 	else
-	{
-	    errmsg = e_invalid_argument;
-	    break;
-	}
-	if (*p == ',')
-	    ++p;
+	    return FALSE;
     }
-    if (errmsg == NULL)
-    {
-	if (global_busy)
-	    // clip_unnamed will be reset to clip_unnamed_saved
-	    // at end_global_changes
-	    clip_unnamed_saved = new_unnamed;
-	else
-	    clip_unnamed = new_unnamed;
-	clip_autoselect_star = new_autoselect_star;
-	clip_autoselect_plus = new_autoselect_plus;
-	clip_autoselectml = new_autoselectml;
-	clip_html = new_html;
-	vim_regfree(clip_exclude_prog);
-	clip_exclude_prog = new_exclude_prog;
-#ifdef FEAT_GUI_GTK
-	if (gui.in_use)
-	{
-	    gui_gtk_set_selection_targets((GdkAtom)GDK_SELECTION_PRIMARY);
-	    gui_gtk_set_selection_targets((GdkAtom)clip_plus.gtk_sel_atom);
-	    gui_gtk_set_dnd_targets();
-	}
+#else
+    return TRUE;
 #endif
-    }
-    else
-	vim_regfree(new_exclude_prog);
-
-    return errmsg;
 }
+#endif
 
 /*
  * Stuff for the X clipboard.  Shared between VMS and Unix.
@@ -2292,18 +2150,10 @@ clip_get_selection(Clipboard_T *cbd)
     }
     else if (!is_clipboard_needs_update())
     {
-#ifdef FEAT_CLIPBOARD_PROVIDER
-	// We will choose if we want to the free the selection if using provider
-	if (clipmethod != CLIPMETHOD_PROVIDER)
-	    clip_free_selection(cbd);
-#endif
+	clip_free_selection(cbd);
 
 	// Try to get selected text from another window
 	clip_gen_request_selection(cbd);
-#ifdef FEAT_CLIPBOARD_PROVIDER
-	if (clipmethod == CLIPMETHOD_PROVIDER)
-	    clip_access_type = CLIP_ACCESS_IMPLICIT;
-#endif
     }
 }
 
@@ -2446,32 +2296,6 @@ may_set_selection(void)
     }
 }
 
-/*
- * Adjust the register name pointed to with "rp" for the clipboard being
- * used always and the clipboard being available.
- */
-    void
-adjust_clip_reg(int *rp)
-{
-    // If no reg. specified, and "unnamed" or "unnamedplus" is in 'clipboard',
-    // use '*' or '+' reg, respectively. "unnamedplus" prevails.
-    if (*rp == 0 && (clip_unnamed != 0 || clip_unnamed_saved != 0))
-    {
-	if (clip_unnamed != 0)
-	    *rp = ((clip_unnamed & CLIP_UNNAMED_PLUS) && clip_plus.available)
-								  ? '+' : '*';
-	else
-	    *rp = ((clip_unnamed_saved & CLIP_UNNAMED_PLUS)
-					   && clip_plus.available) ? '+' : '*';
-    }
-    if ((!clip_star.available && *rp == '*') ||
-	   (!clip_plus.available && *rp == '+'))
-    {
-	msg_warn_missing_clipboard(!clip_plus.available, !clip_star.available);
-	*rp = 0;
-    }
-}
-
 #if defined(FEAT_WAYLAND_CLIPBOARD)
 
     static clip_wl_selection_T *
@@ -2913,6 +2737,8 @@ clip_init_wayland(void)
     if (wayland_ct == NULL)
 	return FAIL;
 
+    memset(&clip_wl, 0, sizeof(clip_wl));
+
     clip_wl.seat = vwl_connection_get_seat(wayland_ct, (char *)p_wse);
 
     if (clip_wl.seat == NULL)
@@ -2930,12 +2756,14 @@ clip_init_wayland(void)
 	    clip_wl.regular.available = true;
 	else
 	{
+	    // Shouldn't happen
 	    vwl_data_device_manager_discard(clip_wl.regular.manager);
 	    clip_wl.regular.manager = NULL;
+	    return FAIL;
 	}
     }
 
-    // If we still don't support the primary selection, find one for it
+    // If we still don't support the primary selection, try finding one for it
     // specifically.
     if (!(supported & WAYLAND_SELECTION_PRIMARY))
     {
@@ -2956,10 +2784,12 @@ clip_init_wayland(void)
 	    }
 	}
     }
-    else if (clip_wl.regular.available)
+
+    if (clip_wl.regular.available && !clip_wl.primary.available)
     {
 	// The protocol supports both regular and primary selections, just use
-	// one data device manager and one data device.
+	// one data device manager and one data device. Or the primary selection
+	// is not supported, make it point to the regular selection instead.
 	clip_wl.primary.available = true;
 	clip_wl.primary.manager = clip_wl.regular.manager;
 	clip_wl.primary.device = clip_wl.regular.device;
@@ -3468,13 +3298,16 @@ clip_wl_owner_exists(Clipboard_T *cbd)
 
 #endif // FEAT_WAYLAND_CLIPBOARD
 
+#endif // FEAT_CLIPBOARD
+
+#ifdef HAVE_CLIPMETHOD
 
 /*
  * Returns the first method for accessing the clipboard that is available/works,
  * depending on the order of values in str.
  */
     static clipmethod_T
-get_clipmethod(char_u *str, bool *plus UNUSED, bool *star)
+get_clipmethod(char_u *str)
 {
     int		len	= (int)STRLEN(str) + 1;
     char_u	*buf	= alloc(len);
@@ -3499,11 +3332,7 @@ get_clipmethod(char_u *str, bool *plus UNUSED, bool *star)
 	    {
 #ifdef FEAT_WAYLAND_CLIPBOARD
 		if (clip_wl.regular.available || clip_wl.primary.available)
-		{
 		    method = CLIPMETHOD_WAYLAND;
-		    *plus = clip_wl.regular.available;
-		    *star = clip_wl.primary.available;
-		}
 #endif
 	    }
 	}
@@ -3514,76 +3343,48 @@ get_clipmethod(char_u *str, bool *plus UNUSED, bool *star)
 #endif
 	    {
 #ifdef FEAT_XCLIPBOARD
-		// x_IOerror_handler() in os_unix.c should set xterm_dpy to NULL
-		// if we lost connection to the X server.
+		// x_IOerror_handler() in os_unix.c should set xterm_dpy to NULL if
+		// we lost connection to the X server.
 		if (xterm_dpy != NULL)
 		{
-		    // If the X connection is lost then that handler will
-		    // longjmp somewhere else, in that case we will call
-		    // choose_clipmethod() again from there, and this if block
-		    // won't be executed since xterm_dpy will be set to NULL.
+		    // If the X connection is lost then that handler will longjmp
+		    // somewhere else, in that case we will call choose_clipmethod()
+		    // again from there, and this if block won't be executed since
+		    // xterm_dpy will be set to NULL.
 		    xterm_update();
 		    method = CLIPMETHOD_X11;
-		    *plus = *star = true;
 		}
 #endif
 	    }
 	}
-	else if (STRCMP(buf, "gui") == 0)
-	{
-#ifdef FEAT_GUI
-	    if (gui.in_use)
-	    {
-		method = CLIPMETHOD_GUI;
-		*star = *plus = true;
-	    }
-#endif
-	}
-	else if (STRCMP(buf, "other") == 0)
-	{
-#if (!defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD) \
-	&& !defined(FEAT_CLIPBOARD_PROVIDER)) \
-	|| (!(defined(UNIX) || defined(MACOS_X)) && defined(FEAT_CLIPBOARD_PROVIDER))
-		method = CLIPMETHOD_OTHER;
-		*plus = *star = true;
-#endif
-	}
 	else
 	{
 #ifdef FEAT_CLIPBOARD_PROVIDER
-	    // Check if it is the name of a provider
-#ifndef ONE_CLIPBOARD
-	    int plus_avail = clip_provider_is_available(&clip_plus, buf);
-#endif
-	    int star_avail = clip_provider_is_available(&clip_star, buf);
+	    // Check if name matches a clipboard provider
+	    int r = clip_provider_is_available(buf);
 
-	    if (
-#ifndef ONE_CLIPBOARD
-		    plus_avail == 1 ||
-#endif
-		    star_avail == 1)
+	    if (r == 1)
 	    {
 		method = CLIPMETHOD_PROVIDER;
-
-		vim_free(clipprovider_name);
-		clipprovider_name = vim_strsave(buf);
-
-#ifndef ONE_CLIPBOARD
-		*plus = plus_avail == 1;
-#endif
-		*star = star_avail == 1;
+		if (ret == CLIPMETHOD_FAIL)
+		{
+		    vim_free(clip_provider);
+		    clip_provider = vim_strsave(buf);
+		    if (clip_provider == NULL)
+			goto fail;
+		}
 	    }
-	    else if (
-#ifndef ONE_CLIPBOARD
-		    plus_avail == -1 ||
-#endif
-		    star_avail == -1)
+	    else if (r == -1)
 #endif
 	    {
+#ifdef FEAT_CLIPBOARD_PROVIDER
+fail:
+#endif
 		ret = CLIPMETHOD_FAIL;
 		goto exit;
 	    }
 	}
+
 	// Keep on going in order to catch errors
 	if (method != CLIPMETHOD_NONE && ret == CLIPMETHOD_FAIL)
 	    ret = method;
@@ -3610,18 +3411,9 @@ clipmethod_to_str(clipmethod_T method)
 	    return (char_u *)"wayland";
 	case CLIPMETHOD_X11:
 	    return (char_u *)"x11";
-	case CLIPMETHOD_GUI:
-	    return (char_u *)"gui";
-	case CLIPMETHOD_OTHER:
-	    return (char_u *)"other";
 	case CLIPMETHOD_PROVIDER:
-#ifdef FEAT_CLIPBOARD_PROVIDER
-	    if (clipprovider_name == NULL)
-		return (char_u *)"none";
-	    else
-		return clipprovider_name;
-#else
-	    return (char_u *)"none";
+#ifdef FEAT_EVAL
+	    return clip_provider;
 #endif
 	default:
 	    return (char_u *)"none";
@@ -3635,61 +3427,81 @@ clipmethod_to_str(clipmethod_T method)
     char *
 choose_clipmethod(void)
 {
-    bool plus = false;
-    bool star = false;
-    clipmethod_T method = get_clipmethod(p_cpm, &plus, &star);
+    clipmethod_T method = get_clipmethod(p_cpm);
 
     if (method == CLIPMETHOD_FAIL)
 	return e_invalid_argument;
 
-#if defined(FEAT_GUI) && defined(FEAT_WAYLAND)
-    if (method == CLIPMETHOD_GUI)
+// If GUI is running or we are not on a system with Wayland or X11, then always
+// return CLIPMETHOD_NONE. System or GUI clipboard handling always overrides.
+// This is unless a provider is being used.
+#if defined(FEAT_XCLIPBOARD) || defined(FEAT_WAYLAND_CLIPBOARD)
+# if defined(FEAT_GUI)
+    if (method != CLIPMETHOD_PROVIDER && gui.in_use)
+    {
+#  ifdef FEAT_WAYLAND
 	// We only interact with Wayland for the clipboard, we can just deinit
 	// everything.
 	wayland_uninit_connection();
+#  endif
+
+	method = CLIPMETHOD_NONE;
+	goto lose_sel_exit;
+    }
+# endif
+#else
+    // If on a system like windows or macos, then clipmethod is irrelevant, we
+    // use their way of accessing the clipboard. This is unless we are using the
+    // clipboard provider
+#ifdef FEAT_CLIPBOARD_PROVIDER
+    if (method != CLIPMETHOD_PROVIDER)
+#endif
+    {
+	method = CLIPMETHOD_NONE;
+	goto exit;
+    }
 #endif
 
+#ifdef FEAT_CLIPBOARD
     // Deinitialize clipboard if there is no way to access clipboard
     if (method == CLIPMETHOD_NONE)
 	clip_init(FALSE);
     // If we have a clipmethod that works now, then initialize clipboard
     else if (clipmethod == CLIPMETHOD_NONE && method != CLIPMETHOD_NONE)
     {
-#ifndef ONE_CLIPBOARD
-	clip_init_single(&clip_plus, plus);
-	clip_plus.did_warn = false;
-#endif
-	clip_init_single(&clip_star, star);
-	clip_star.did_warn = false;
+	clip_init(TRUE);
+	did_warn_clipboard = false;
     }
-    else if ((clipmethod != CLIPMETHOD_NONE && method != clipmethod))
+    // Disown clipboard if we are switching to a new method
+    else if (clipmethod != CLIPMETHOD_NONE && method != clipmethod)
     {
-	// Disown clipboard if we are switching to a new method
+#if (defined(FEAT_XCLIPBOARD) || defined(FEAT_WAYLAND_CLIPBOARD)) \
+	&& defined(FEAT_GUI)
+lose_sel_exit:
+#endif
 	if (clip_star.owned)
 	    clip_lose_selection(&clip_star);
-	clip_init_single(&clip_star, star);
-#ifndef ONE_CLIPBOARD
 	if (clip_plus.owned)
 	    clip_lose_selection(&clip_plus);
-	clip_init_single(&clip_plus, plus);
+
+#if defined(FEAT_GUI)
+	if (!gui.in_use)
 #endif
+	{
+	    clip_init(TRUE);
+	    did_warn_clipboard = false;
+	}
     }
-    else
-    {
-	// If availability of a clipboard changed, then update the clipboard
-	// structure.
-#ifndef ONE_CLIPBOARD
-	if (plus != clip_plus.available)
-	    clip_init_single(&clip_plus, plus);
+#endif // FEAT_CLIPBOARD
+
+#if !defined(FEAT_XCLIPBOARD) && !defined(FEAT_WAYLAND_CLIPBOARD)
+exit:
 #endif
-	if (star != clip_star.available)
-	    clip_init_single(&clip_star, star);
-    }
 
     clipmethod = method;
 
 #ifdef FEAT_EVAL
-    set_vim_var_string(VV_CLIPMETHOD, clipmethod_to_str(method), -1);
+    set_vim_var_string(VV_CLIPMETHOD, (char_u*)clipmethod_to_str(method), -1);
 #endif
 
     return NULL;
@@ -3712,33 +3524,175 @@ ex_clipreset(exarg_T *eap UNUSED)
 		clipmethod_to_str(clipmethod));
 }
 
+/*
+ * Adjust the register name pointed to with "rp" for the clipboard being
+ * used always and the clipboard being available.
+ */
+    void
+adjust_clip_reg(int *rp)
+{
+#ifdef FEAT_CLIPBOARD_PROVIDER
+    if (clipmethod == CLIPMETHOD_PROVIDER)
+    {
+	if (clip_unnamed != 0)
+	    *rp = ((clip_unnamed & CLIP_UNNAMED_PLUS)) ? '+' : '*';
+	return;
+    }
+#endif
+#ifdef FEAT_CLIPBOARD
+    // If no reg. specified, and "unnamed" or "unnamedplus" is in 'clipboard',
+    // use '*' or '+' reg, respectively. "unnamedplus" prevails.
+    if (*rp == 0 && (clip_unnamed != 0 || clip_unnamed_saved != 0))
+    {
+	if (clip_unnamed != 0)
+	    *rp = ((clip_unnamed & CLIP_UNNAMED_PLUS) && clip_plus.available)
+								  ? '+' : '*';
+	else
+	    *rp = ((clip_unnamed_saved & CLIP_UNNAMED_PLUS)
+					   && clip_plus.available) ? '+' : '*';
+    }
+    if ((!clip_star.available && *rp == '*') ||
+	   (!clip_plus.available && *rp == '+'))
+    {
+	msg_warn_missing_clipboard();
+	*rp = 0;
+    }
+#endif
+}
+
+/*
+ * Extract the items in the 'clipboard' option and set global values.
+ * Return an error message or NULL for success.
+ */
+    char *
+did_set_clipboard(optset_T *args UNUSED)
+{
+    int		new_unnamed = 0;
+#ifdef FEAT_CLIPBOARD
+    int		new_autoselect_star = FALSE;
+    int		new_autoselect_plus = FALSE;
+    int		new_autoselectml = FALSE;
+    int		new_html = FALSE;
+#endif
+    regprog_T	*new_exclude_prog = NULL;
+    char	*errmsg = NULL;
+    char_u	*p;
+
+    for (p = p_cb; *p != NUL; )
+    {
+	// Note: Keep this in sync with p_cb_values.
+	if (STRNCMP(p, "unnamed", 7) == 0 && (p[7] == ',' || p[7] == NUL))
+	{
+	    new_unnamed |= CLIP_UNNAMED;
+	    p += 7;
+	}
+	else if (STRNCMP(p, "unnamedplus", 11) == 0
+					    && (p[11] == ',' || p[11] == NUL))
+	{
+	    new_unnamed |= CLIP_UNNAMED_PLUS;
+	    p += 11;
+	}
+#ifdef FEAT_CLIPBOARD
+	else if (STRNCMP(p, "autoselect", 10) == 0
+					    && (p[10] == ',' || p[10] == NUL))
+	{
+	    new_autoselect_star = TRUE;
+	    p += 10;
+	}
+	else if (STRNCMP(p, "autoselectplus", 14) == 0
+					    && (p[14] == ',' || p[14] == NUL))
+	{
+	    new_autoselect_plus = TRUE;
+	    p += 14;
+	}
+	else if (STRNCMP(p, "autoselectml", 12) == 0
+					    && (p[12] == ',' || p[12] == NUL))
+	{
+	    new_autoselectml = TRUE;
+	    p += 12;
+	}
+	else if (STRNCMP(p, "html", 4) == 0 && (p[4] == ',' || p[4] == NUL))
+	{
+	    new_html = TRUE;
+	    p += 4;
+	}
+	else if (STRNCMP(p, "exclude:", 8) == 0 && new_exclude_prog == NULL)
+	{
+	    p += 8;
+	    new_exclude_prog = vim_regcomp(p, RE_MAGIC);
+	    if (new_exclude_prog == NULL)
+		errmsg = e_invalid_argument;
+	    break;
+	}
+#endif
+	else
+	{
+	    errmsg = e_invalid_argument;
+	    break;
+	}
+	if (*p == ',')
+	    ++p;
+    }
+    if (errmsg == NULL)
+    {
+#ifdef FEAT_CLIPBOARD
+	if (global_busy)
+	    // clip_unnamed will be reset to clip_unnamed_saved
+	    // at end_global_changes
+	    clip_unnamed_saved = new_unnamed;
+	else
+#endif
+	    clip_unnamed = new_unnamed;
+#ifdef FEAT_CLIPBOARD
+	clip_autoselect_star = new_autoselect_star;
+	clip_autoselect_plus = new_autoselect_plus;
+	clip_autoselectml = new_autoselectml;
+	clip_html = new_html;
+	vim_regfree(clip_exclude_prog);
+	clip_exclude_prog = new_exclude_prog;
+#endif
+#ifdef FEAT_GUI_GTK
+	if (gui.in_use)
+	{
+	    gui_gtk_set_selection_targets((GdkAtom)GDK_SELECTION_PRIMARY);
+	    gui_gtk_set_selection_targets((GdkAtom)clip_plus.gtk_sel_atom);
+	    gui_gtk_set_dnd_targets();
+	}
+#endif
+    }
+    else
+	vim_regfree(new_exclude_prog);
+
+    return errmsg;
+}
+
+#endif // HAVE_CLIPMETHOD
+
 #ifdef FEAT_CLIPBOARD_PROVIDER
 
 /*
- * Check if a clipboard provider with given name exists and is available for the
- * given clipboard. Returns 1 if the provider exists and the 'available'
- * function returned true, 0 if the provider exists but the function returned
- * false, and -1 on error.
+ * Check if a clipboard provider with given name is available. Returns 1 if available,
+ * 0 if not available, and -1 on error
  */
     static int
-clip_provider_is_available(Clipboard_T *cbd, char_u *provider)
+clip_provider_is_available(char_u *provider)
 {
     dict_T	*providers = get_vim_var_dict(VV_CLIPPROVIDERS);
     typval_T	provider_tv = {0};
     callback_T	callback = {0};
     typval_T	rettv = {0};
     typval_T	func_tv = {0};
-    char_u	*avail;
     int		res = 0;
 
     if (dict_get_tv(providers, (char *)provider, &provider_tv) == FAIL
 	    || provider_tv.v_type != VAR_DICT)
+	// clipboard provider not defined
 	return -1;
 
     if (dict_get_tv(provider_tv.vval.v_dict, "available", &func_tv) == FAIL)
     {
 	clear_tv(&provider_tv);
-	// If "available" functon not specified assume always TRUE
+	// If "available" function not specified assume always TRUE
 	return 1;
     }
 
@@ -3746,17 +3700,10 @@ clip_provider_is_available(Clipboard_T *cbd, char_u *provider)
 	goto fail;
 
     if (call_callback(&callback, -1, &rettv, 0, NULL) == FAIL ||
-	    rettv.v_type != VAR_STRING)
+	    (rettv.v_type != VAR_BOOL && rettv.v_type != VAR_NUMBER))
 	goto fail;
 
-    avail = rettv.vval.v_string;
-
-
-    if (
-#ifndef ONE_CLIPBOARD
-	    (vim_strchr(avail, '+') != NULL && cbd == &clip_plus) ||
-#endif
-	    (vim_strchr(avail, '*') != NULL && cbd == &clip_star))
+    if (rettv.vval.v_number)
 	res = 1;
 
     if (FALSE)
@@ -3772,7 +3719,7 @@ fail:
 }
 
 /*
- * Get the specified callback "function" from the provider dictionary of for
+ * Get the specified callback "function" from the provider dictionary for
  * register "reg".
  */
     static int
@@ -3833,9 +3780,8 @@ clip_provider_get_callback(
 }
 
     static void
-clip_provider_set_selection(Clipboard_T *cbd, char_u *provider)
+clip_provider_copy(char_u *reg, char_u *provider)
 {
-    char_u	*reg = (char_u *)(cbd == &clip_star ? "*" : "+");
     callback_T	callback;
     typval_T	rettv;
     typval_T	argvars[4];
@@ -3850,14 +3796,9 @@ clip_provider_set_selection(Clipboard_T *cbd, char_u *provider)
 		&callback) == FAIL)
 	return;
 
-    // Possibly get selected text, if using autoselect for 'clipboard'
-    cbd->owned = TRUE;
-    clip_get_selection(cbd);
-    cbd->owned = FALSE;
-
     // Convert register type into a string
-    if (cbd == &clip_plus)
-	y_ptr = get_y_register(PLUS_REGISTER);
+    if (*reg == '+')
+	y_ptr = get_y_register(REAL_PLUS_REGISTER);
     else
 	y_ptr = get_y_register(STAR_REGISTER);
 
@@ -3918,11 +3859,10 @@ clip_provider_set_selection(Clipboard_T *cbd, char_u *provider)
 }
 
     static void
-clip_provider_request_selection(Clipboard_T *cbd, char_u *provider)
+clip_provider_paste(char_u *reg, char_u *provider)
 {
-    char_u	*reg = (char_u *)(cbd == &clip_star ? "*" : "+");
     callback_T	callback;
-    typval_T	argvars[3];
+    typval_T	argvars[2];
     typval_T	rettv;
     int		ret;
     char_u	*reg_type;
@@ -3938,27 +3878,14 @@ clip_provider_request_selection(Clipboard_T *cbd, char_u *provider)
     argvars[0].v_type = VAR_STRING;
     argvars[0].vval.v_string = reg;
 
-    argvars[1].v_type = VAR_STRING;
-    argvars[1].vval.v_string = (char_u *)
-	(clip_access_type == CLIP_ACCESS_EXPLICIT ? "explicit" : "implicit");
-
-    argvars[2].v_type = VAR_UNKNOWN;
+    argvars[1].v_type = VAR_UNKNOWN;
 
     textlock++;
-    ret = call_callback(&callback, -1, &rettv, 2, argvars);
+    ret = call_callback(&callback, -1, &rettv, 1, argvars);
     textlock--;
 
     if (ret == FAIL)
 	goto exit;
-    else if (rettv.v_type == VAR_STRING
-	    && STRCMP(rettv.vval.v_string, "clear") == 0)
-    {
-	clip_free_selection(cbd);
-	goto exit;
-    }
-    else if (rettv.v_type == VAR_STRING
-	    && STRCMP(rettv.vval.v_string, "previous") == 0)
-	goto exit;
     else if (rettv.v_type == VAR_TUPLE
 	    && TUPLE_LEN(rettv.vval.v_tuple) == 2
 	    && TUPLE_ITEM(rettv.vval.v_tuple, 0)->v_type == VAR_STRING
@@ -3979,51 +3906,84 @@ clip_provider_request_selection(Clipboard_T *cbd, char_u *provider)
 	goto exit;
 
     {
-	char_u	    yank_type = MAUTO;
-	long	    block_len = -1;
-	yankreg_T   *y_ptr;
-	char_u	    **contents;
-	listitem_T  *li;
-	int	    i = 0;
+	char_u		yank_type = MAUTO;
+	long		block_len = -1;
+	yankreg_T	*y_ptr, *cur_y_ptr;
+	char_u		**lstval;
+	char_u		**allocval;
+	char_u		buf[NUMBUFLEN];
+	char_u		**curval;
+	char_u		**curallocval;
+	char_u		*strval;
+	listitem_T	*li;
+	int		len;
 
-	contents = ALLOC_MULT(char_u *, lines->lv_len + 1); // Ends with a NULL
+	// If the list is NULL handle like an empty list.
+	len = lines == NULL ? 0 : lines->lv_len;
 
-	if (contents == NULL)
+	// First half: use for pointers to result lines; second half: use for
+	// pointers to allocated copies.
+	lstval = ALLOC_MULT(char_u *, (len + 1) * 2);
+	if (lstval == NULL)
 	    goto exit;
+	curval = lstval;
+	allocval = lstval + len + 2;
+	curallocval = allocval;
 
-	// Convert strings in list to type char_u **
-	FOR_ALL_LIST_ITEMS(lines, li)
+	if (lines != NULL)
 	{
-	    char_u *str = tv_get_string_chk(&li->li_tv);
-
-	    if (str == NULL)
-		goto exit;
-
-	    contents[i++] = vim_strsave(str);
+	    CHECK_LIST_MATERIALIZE(lines);
+	    FOR_ALL_LIST_ITEMS(lines, li)
+	    {
+		strval = tv_get_string_buf_chk(&li->li_tv, buf);
+		if (strval == NULL)
+		    goto free_lstval;
+		if (strval == buf)
+		{
+		    // Need to make a copy, next tv_get_string_buf_chk() will
+		    // overwrite the string.
+		    strval = vim_strsave(buf);
+		    if (strval == NULL)
+			goto free_lstval;
+		    *curallocval++ = strval;
+		}
+		*curval++ = strval;
+	    }
 	}
-	contents[i] = NULL;
+	*curval++ = NULL;
 
-	if (STRLEN(reg_type) > 0
-		&& get_yank_type(&reg_type, &yank_type, &block_len) == FAIL)
-	    goto exit;
+	if (*reg_type != NUL && (STRLEN(reg_type) <= 0
+		|| get_yank_type(&reg_type, &yank_type, &block_len) == FAIL))
+	{
+	    emsg(e_invalid_argument);
+	    goto free_lstval;
+	}
 
-	if (cbd == &clip_plus)
-	    y_ptr = get_y_register(PLUS_REGISTER);
+	if (*reg == '+')
+	    y_ptr = get_y_register(REAL_PLUS_REGISTER);
 	else
 	    y_ptr = get_y_register(STAR_REGISTER);
 
-	clip_free_selection(cbd);
+	// Free previous register contents
+	cur_y_ptr = get_y_current();
+	set_y_current(y_ptr);
+
+	free_yank_all();
+	get_y_current()->y_size = 0;
+
+	set_y_current(cur_y_ptr);
 
 	str_to_reg(y_ptr,
 		yank_type,
-		(char_u *)contents,
-		STRLEN(contents),
+		(char_u *)lstval,
+		-1,
 		block_len,
 		TRUE);
 
-	for (int k = 0; k < i; k++)
-	    vim_free(contents[k]);
-	vim_free(contents);
+free_lstval:
+	while (curallocval > allocval)
+	    vim_free(*--curallocval);
+	vim_free(lstval);
     }
 
 exit:
@@ -4031,6 +3991,82 @@ exit:
     clear_tv(&rettv);
 }
 
-#endif // FEAT_CLIPBOARD_PROVIDER
+// Used to stop calling the provider callback every time there is an update.
+// This prevents unnecessary calls when accessing the provider often in an
+// interval.
+//
+// If -1 then allow provider callback to be called then set to one. Default
+// value (is allowed) is -2.
+static int star_pause_count = -2, plus_pause_count = -2;
 
-#endif // FEAT_CLIPBOARD
+    void
+call_clip_provider_request(int reg)
+{
+    if (clipmethod != CLIPMETHOD_PROVIDER)
+	return;
+
+    if (reg == '+' && plus_pause_count < 0)
+    {
+	if (plus_pause_count == -1)
+	    plus_pause_count = 1;
+	clip_provider_paste((char_u *)"+", clip_provider);
+    }
+    else if (reg == '*' && star_pause_count < 0)
+    {
+	if (star_pause_count == -1)
+	    star_pause_count = 1;
+	clip_provider_paste((char_u *)"*", clip_provider);
+    }
+    else
+	return;
+}
+
+    void
+call_clip_provider_set(int reg)
+{
+    if (clipmethod != CLIPMETHOD_PROVIDER)
+	return;
+
+    if (reg == '+' && plus_pause_count < 0)
+    {
+	if (plus_pause_count == -1)
+	    plus_pause_count = 1;
+	clip_provider_copy((char_u *)"+", clip_provider);
+    }
+    else if (reg == '*' && star_pause_count < 0)
+    {
+	if (star_pause_count == -1)
+	    star_pause_count = 1;
+	clip_provider_copy((char_u *)"*", clip_provider);
+    }
+}
+
+/*
+ * Makes it so that the next provider call is only done once any calls after are
+ * ignored, until dec_clip_provider is called the same number of times after
+ * again. Note that this is per clipboard register ("+", "*")
+ */
+    void
+inc_clip_provider(void)
+{
+    plus_pause_count = (plus_pause_count == -2
+	|| plus_pause_count == -1) ? -1 : plus_pause_count + 1;
+    star_pause_count = (star_pause_count == -2
+	|| star_pause_count == -1) ? -1 : star_pause_count + 1;
+}
+
+    void
+dec_clip_provider(void)
+{
+    if (plus_pause_count != -2)
+	plus_pause_count = plus_pause_count == -1 ? -1 : plus_pause_count - 1;
+    if (star_pause_count != -2)
+	star_pause_count = star_pause_count == -1 ? -1 : star_pause_count - 1;
+
+    if (plus_pause_count == 0 || plus_pause_count == -1)
+	plus_pause_count = -2;
+    if (star_pause_count == 0 || star_pause_count == -1)
+	star_pause_count = -2;
+}
+
+#endif // FEAT_CLIPBOARD_PROVIDER
